/*

  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/

  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of authors nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.

*/

#if HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef DEBUG
#define NDEBUG
#endif
#include <stdio.h>
#if HAVE_TERM_H
#include <term.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_ASSERT_H
#include <assert.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_CTYPE_H
#include <ctype.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif

#include "uim-fep.h"
#include "draw.h"
#include "escseq.h"
#include "str.h"
#include "read.h"

/* ä */
static int s_use_civis = FALSE;
static int s_status;
static int s_init = FALSE;
/* ߤΥ */
static struct point_tag s_cursor = {UNDEFINED, UNDEFINED};
static struct point_tag s_cursor_diff = {0, 0};

/* 뤬äƤȤTRUE */
static int s_cursor_invisible = FALSE;
/* ֤¸ȤTRUEˤ */
static int s_save = FALSE;
static struct point_tag s_save_cursor;
/* TRUEΤȤ°ѤäƤǽ */
static int s_attribute = FALSE;
/* TRUEΤȤ⡼ */
static int s_underline = FALSE;
/* TRUEΤȤȿž⡼ */
static int s_standout = FALSE;

static void fixtty(void);

/*
 * termcap/terminfoΥȥ꤬뤫ǧ
 */
void init_escseq(int use_civis, int statusline_type)
{
  s_use_civis = use_civis;
  s_status = statusline_type;

  if (enter_underline_mode == NULL) {
    printf("enter_underline_mode is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (exit_underline_mode == NULL) {
    printf("exit_underline_mode is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (enter_standout_mode == NULL) {
    printf("enter_standout_mode is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (exit_standout_mode == NULL) {
    printf("exit_standout_mode is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (exit_attribute_mode == NULL) {
    printf("exit_attribute_mode is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (clr_eol == NULL) {
    printf("clr_eol is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (cursor_address == NULL) {
    printf("cursor_address is not available\r\n");
    done(EXIT_FAILURE);
  }
  if (s_status == LASTLINE) {
    if (change_scroll_region == NULL) {
      printf("change_scroll_region is not available\r\n");
      done(EXIT_FAILURE);
    }
    /*
    if (clear_screen == NULL) {
      printf("clear_screen is not available\r\n");
      done(EXIT_FAILURE);
    }
    if (clr_eos == NULL) {
      printf("clr_eos is not available\r\n");
      done(EXIT_FAILURE);
    }
    */
  }
  fixtty();
  s_init = TRUE;
}

void quit_escseq(void)
{
  if (s_init == FALSE) {
    return;
  }
  if (s_status == LASTLINE) {
    put_save_cursor();
    put_change_scroll_region(0, g_win->ws_row);
    clear_lastline();
    put_restore_cursor();
  }
  put_exit_attribute_mode();
  put_exit_standout_mode();
  put_exit_underline_mode();
  put_restore_cursor();
  put_cursor_normal();
}

static void fixtty(void)
{
  struct point_tag start_cursor;
  struct point_tag cursor;
  struct point_tag cursor2;
  static struct termios tios;
  tcgetattr(STDIN_FILENO, &tios);
  cfmakeraw(&tios);
  /* read_stdinޤǤɤޤʤФʤʤǾʸ */
  tios.c_cc[VMIN] = 0;
  /* read_stdinޤǤΥॢ 0.1ñ */
  tios.c_cc[VTIME] = 3;
  tcsetattr(STDIN_FILENO, TCSAFLUSH, &tios);
  /* ϰ֤¸ */
  start_cursor = get_cursor_position();
  if (start_cursor.row == UNDEFINED) {
    printf("Report Cursor Position is not available\r\n");
    done(EXIT_FAILURE);
  }
  put_cursor_invisible();
  /* ǲԤ鳫ϤȤΤ˥ */
  if (s_status == LASTLINE) {
    putchar('\n');
  }
  /* ʰ֤˰ư */
  put_cursor_address(1, 1);
  s_cursor.row = s_cursor.col = UNDEFINED;
  /* ֤ */
  cursor = get_cursor_position();
  s_cursor.row = s_cursor.col = UNDEFINED;
  /* ֤˰ư */
  put_cursor_address_p(&cursor);
  s_cursor.row = s_cursor.col = UNDEFINED;
  /* Ʊ֤뤫 */
  cursor2 = get_cursor_position();
  /* ʤä麹ʬĴ٤ */
  s_cursor_diff.row = cursor2.row - cursor.row;
  s_cursor_diff.col = cursor2.col - cursor.col;
  start_cursor.row -= s_cursor_diff.row;
  start_cursor.col -= s_cursor_diff.col;
  if (s_status == LASTLINE) {
    put_change_scroll_region(0, g_win->ws_row - 1);
  }
  if (s_status != BACKTICK) {
    draw_statusline_no_restore();
  }
  /* ϰ֤ */
  put_cursor_address_p(&start_cursor);
  put_cursor_normal();
}

/*
 * ֤¸
 * put_restore_cursor()ȸߤ˸Ƥ
 */
void put_save_cursor(void)
{
  if (!s_save) {
    debug(("<put_save_cursor>"));
    s_save_cursor = get_cursor_position();
    s_save = TRUE;
  }
}

/*
 * ¸֤
 * put_save_cursor()ȸߤ˸Ƥ
 */
void put_restore_cursor(void)
{
  if (s_save) {
    debug(("<put_restore_cursor>"));
    put_cursor_address_p(&s_save_cursor);
    s_save = FALSE;
  }
}

/*
 * ߤΥ֤֤
 * 0, 0
 * ֤ʤäUNDEFINED֤
 */
struct point_tag get_cursor_position(void)
{
  char ibuf[100];
  char *ibuf_offset;
  ssize_t len;
  char *escseq = NULL;
  int loop_count = 0;
  char *unget_buf = NULL;
  char *unget_buf2 = NULL;
  int buf_size = 0;
  int offset = 0;
  if (s_cursor.row != UNDEFINED) {
    return s_cursor;
  }
  printf("\033[6n");

  do {
    /* 10롼פƤä齪λ */
    if (loop_count++ >= 10) {
      break;
    }
    ibuf_offset = ibuf + offset;
    len = read_stdin(ibuf_offset, sizeof(ibuf) - offset);
    if (len <= 0) {
      continue;
    }
    ibuf_offset[len] = '\0';
    debug(("get = %s\n", ibuf));
    escseq = strstr(ibuf, "\033");
    if (escseq != NULL) {
      int n;
      unget_buf2 = malloc(offset + len);
      /* unget_buf2[0] == 'R' */
      n = sscanf(escseq, "\033[%d;%d%s", &(s_cursor.row), &(s_cursor.col), unget_buf2);
      if (n < 3) {
        /* ESCϤ뤬, ץ󥹤ȤƤԽʬ */
        offset += len;
        s_cursor.row = s_cursor.col = UNDEFINED;
      } else {
        char *R = strchr(escseq, 'R');
        /* Rɬ */
        assert(R != NULL);
        /* Rμʸ뤫 */
        if (R[1] != '\0') {
          strcpy(unget_buf2, R + 1);
        } else {
          free(unget_buf2);
          unget_buf2 = NULL;
        }
      }
    } else {
      offset += len;
    }
  } while (s_cursor.row == UNDEFINED);

  /* ץ󥹤ʸ󤬤뤫 */
  if (escseq > ibuf) {
    unget_buf = malloc(escseq - ibuf);
    memcpy(unget_buf, ibuf, escseq - ibuf);
    buf_size += (escseq - ibuf);
  }

  /* ץ󥹤θʸ󤬤뤫 */
  if (unget_buf2 != NULL) {
    int buf_size2 = strlen(unget_buf2);
    unget_buf = realloc(unget_buf, buf_size + buf_size2);
    memcpy(unget_buf + buf_size, unget_buf2, buf_size2);
    buf_size += buf_size2;
    free(unget_buf2);
  }

  if (buf_size > 0) {
    unget_stdin(unget_buf, buf_size);
    free(unget_buf);
  }

  if (s_cursor.row == UNDEFINED) {
    /*  */
    /* s_cursor.row = s_cursor.col = 0; */
    return s_cursor;
  }

  s_cursor.row--;
  s_cursor.col--;
  s_cursor.row -= s_cursor_diff.row;
  s_cursor.col -= s_cursor_diff.col;

  /* screenǤϤʤ뤳Ȥ */
  if (s_cursor.col > g_win->ws_col - 1) {
    put_crlf();
  }
  debug(("<get row = %d col = %d>", s_cursor.row, s_cursor.col));
  return s_cursor;
}

/*
 * 򸫤ʤ褦ˤ
 */
void put_cursor_invisible(void)
{
  if (s_use_civis && !s_cursor_invisible && cursor_invisible != NULL && cursor_normal != NULL) {
    putp(cursor_invisible);
    s_cursor_invisible = TRUE;
    debug(("<invis>"));
  }
}

/*
 * 򸫤褦ˤ
 */
void put_cursor_normal(void)
{
  if (s_cursor_invisible) {
    putp(cursor_normal);
    s_cursor_invisible = FALSE;
    debug(("<norm>"));
  }
}

/*
 * underline⡼ɤ򳫻Ϥ
 */
void put_enter_underline_mode(void)
{
  if (!s_underline) {
    s_underline = TRUE;
    putp(enter_underline_mode);
    debug(("<{_>"));
  }
}

/*
 * underline⡼ɤˤʤäƤ顢underline⡼ɤstandout⡼ɤλ
 */
void put_exit_underline_mode(void)
{
  if (s_underline) {
    s_underline = FALSE;
    putp(exit_underline_mode);
    assert(s_standout == FALSE);
    if (s_standout) {
      putp(exit_standout_mode);
    }
    debug(("<_}>"));
  }
}

/*
 * standout⡼ɤ򳫻Ϥ
 */
void put_enter_standout_mode(void)
{
  if (!s_standout) {
    s_standout = TRUE;
    putp(enter_standout_mode);
    debug(("<{r>"));
  }
}

/*
 * standout⡼ɤˤʤäƤ顢underline⡼ɤstandout⡼ɤλ
 */
void put_exit_standout_mode(void)
{
  if (s_standout) {
    s_standout = FALSE;
    putp(exit_standout_mode);
    assert(s_underline == FALSE);
    if (s_underline) {
      putp(exit_underline_mode);
    }
    debug(("<r}>"));
  }
}

/*
 * standout⡼ɡunderline⡼ɤλ
 * ʸطʿ򸵤᤹
 */
void put_exit_attribute_mode(void)
{
  if (s_attribute) {
    s_attribute = s_underline = s_standout = FALSE;
    putp(exit_attribute_mode);
    debug(("<a}>"));
  }
}

/*
 * exit_attribute_modeϤ
 */
/*
void put_exit_attribute_mode_force()
{
  s_attribute = TRUE;
  put_exit_attribute_mode();
}
*/

/*
 * row col˰ư
 * 0, 0
 */
void put_cursor_address(int row, int col)
{
  if (row >= g_win->ws_row) {
    row = g_win->ws_row - 1;
  }
  /* üؤΰưϾάʤ */
  if (row == s_cursor.row && col == s_cursor.col && col < g_win->ws_col - 2) {
    return;
  }
  putp(tparm(cursor_address, row, col));
  s_cursor.row = row;
  s_cursor.col = col;
  debug(("<go %d %d>", row, col));
}

void put_cursor_address_p(struct point_tag *p)
{
  put_cursor_address(p->row, p->col);
}

/*
 * Ԥ
 */
void put_crlf(void)
{
  debug(("<crlf>"));
  printf("\r\n");
  s_cursor.col = 0;
  s_cursor.row++;
  if (s_cursor.row >= g_win->ws_row) {
    s_cursor.row = g_win->ws_row - 1;
  }
}

/*
 * ǲԤcol˰ư
 */
void put_goto_lastline(int col)
{
  int row = g_win->ws_row;
  if (row == s_cursor.row && col == s_cursor.col) {
    return;
  }
  putp(tparm(cursor_address, row, col));
  s_cursor.row = row;
  s_cursor.col = col;
  debug(("<go %d %d>", row, col));
}

/*
 * underline⡼ɤstandout⡼ɤλnʸڡϤ
 * ̤αüۤƤϤʤ
 */
void put_erase(int n)
{
  int i;
  if (n <= 0) {
    return;
  }
  put_exit_underline_mode();
  put_exit_standout_mode();
  for (i = 0; i < n; i++) {
    printf(" ");
    debug(("<sp>"));
  }
  s_cursor.col += n;
  assert(s_cursor.col <= g_win->ws_col);
}

/*
 * underline⡼ɤstandout⡼ɤλƹޤǾõ
 * ֤Ѥʤ
 */
void put_clear_to_end_of_line(void)
{
  put_exit_standout_mode();
  put_exit_underline_mode();
  putp(clr_eol);
  debug(("<clear>"));
}

/*
 * üΰstartendꤹ
 */
void put_change_scroll_region(int start, int end)
{
  putp(tparm(change_scroll_region, start, end));
  s_cursor.row = s_cursor.col = 0;
  debug(("<region %d %d>", start, end));
}

/*
 * strü˽Ϥ
 * ̤αüۤƤϤʤ
 * ץ󥹤ϴޤޤʤ
 */
void put_str(const char *str)
{
  if (str[0] == '\0') {
    return;
  }
  debug(("<put_str \"%s\">", str));
  s_cursor.col += strwidth(str);
  assert(s_cursor.col <= g_win->ws_col);
  printf("%s", str);
}

/* strlenϤ
 * const char *strȤʤäƤ뤬Ū˽񴹤Τstrʸ
 * ˤƤϤʤ
 */
void put_str_len(const char *str, int len)
{
  char save_char = str[len];
  ((char *)str)[len] = '\0';
  put_str(str);
  ((char *)str)[len] = save_char;
}

/*
 * ptyνstrü˽Ϥ
 * ץ󥹤ޤ礬
 */
void put_str_from_pty(const char *str)
{
  if (str[0] == '\0') {
    return;
  }
  printf("%s", str);
  debug(("<put_str_from_pty \"%s\">", str));
  s_attribute = TRUE;
  g_commit = FALSE;
  s_cursor.row = s_cursor.col = UNDEFINED;
}


/*
 * escseq/$<\d+>/Τ֤
 * escseqNULLäNULL֤
 */
char *cut_padding(const char *escseq)
{
  int i, j;
  char *cut_str;
  if (escseq == NULL) {
    return NULL;
  }
  cut_str = malloc(strlen(escseq) + 1);
  for (i = 0, j = 0; escseq[i] != '\0';) {
    if (escseq[i] == '$' && escseq[i + 1] == '<') {
      int i2 = i + 1;
      while (isdigit(escseq[++i2]));
      if (escseq[i2] == '>') {
        i = i2 + 1;
        if (escseq[i] == '\0') {
          break;
        }
      }
    }
    cut_str[j++] = escseq[i++];
  }
  cut_str[j] = '\0';
  return cut_str;
}

/*
 * üѹ줿Ȥ˸Ƥ
 */
void escseq_winch(void)
{
  s_cursor.row = s_cursor.col = UNDEFINED;
}
